/* Header file for the MTsort system task structure ... */

#include <version.h>
#include <runfile.h>
#include <mtsortstats.h>
#include <errno.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <pthread.h>

#define TRYWAIT_AGAIN         EINTR

#define MAX_INPUT_STREAMS   1
#define MAX_EVENT_STREAMS   1
#define MAX_OUTPUT_STREAMS  4

#define MAX_INPUT_RUNRECORDS  1024
#define MAX_OUTPUT_RUNRECORDS 1024

#define HDR_ID1 0x19062002
#define HDR_ID2 0x09592400

/* 32bit file access, fails on max-1 */

#define MAX_FILE_BLOCK  0x7fffffff
#define MAX_FILE_OFFSET 2147418112     /* =(2147483648 - 65536) */

#define MAX_OUTPUT_FILE_BLOCK 65535     /* 2G/32k -1 */


#define UINT32 unsigned int
#define UINT   unsigned int
#define ULONG  unsigned long
#define STATUS int

#define ERROR      -1
#define OK          0

#define FALSE       0
#define TRUE        1

extern int errno;

/* Handler states ... */

#define UNINITIALISED 0
#define STOPPED       1
#define OPENING       2
#define OPENED        3
#define PAUSED        4
#define GOING         5
#define FLUSHING      6
#define EXITING       7

/* Stream states ... */

#define ST_FREE       0
#define ST_OPEN       1
#define ST_OPEN_EOF   3

/* ... action states */

#define ST_NONE       0
#define ST_SETPOS     1
#define ST_DUMMY1     2
#define ST_DUMMY2     4
#define ST_DUMMY3     8
#define ST_NEWFILE   16
#define ST_REVOKE    32
#define ST_NEWDEVICE 64
#define ST_FLUSH  16384
#define ST_EXIT   32768

/* Buffer states ... */

static char *file_states[]  = {"FREE","FILLING","FULL","CLEARING","OLD","INVALID"};

enum {FREE=0, FILLING, FULL, CLEARING, OLD, INVALID, N_STATES };

/* file close modes ... */

#define MODE_FSF 0
#define MODE_REWIND 1
#define MODE_FREEZE 2

/* Buffer definitions ... */

#define MTSORTSEGMENTSBASE    0x62000000
#define MTSORTLOCALDATABASE   0x63000000     //Was 62400000

#define INPUT_BUFFERS_PER_STREAM    64
#define OUTPUT_BUFFERS_PER_STREAM   4

#define MAX_INPUT_BUFFERS   (INPUT_BUFFERS_PER_STREAM*MAX_INPUT_STREAMS)
#define MAX_EVENT_BUFFERS   MAX_INPUT_BUFFERS
#define MAX_OUTPUT_BUFFERS  (OUTPUT_BUFFERS_PER_STREAM*MAX_OUTPUT_STREAMS)

#define MAX_BUFFER_SIZE  131072
#define DEFAULT_EVENT_BUFFER_SIZE  65536

#define INPUT_BUFFER_ALLOCATED_SIZE  (MAX_BUFFER_SIZE+1024)
#define EVENT_BUFFER_ALLOCATED_SIZE  (MAX_BUFFER_SIZE+1024)
#define OUTPUT_BUFFER_ALLOCATED_SIZE (MAX_BUFFER_SIZE+1024)

#define MAX_DATA_LENGTH  (MAX_BUFFER_SIZE-32)

#define DATA_HEADER_LENGTH    24

/* Input/Output handler types ... (add to list for different variants) */

enum {SCHEDULER=0, INPUTHANDLER, EVENTHANDLER, OUTPUTHANDLER, 
                   INPUTSTATICHANDLER, EVENTSTATICHANDLER, OUTPUTSTATICHANDLER,
	           N_HANDLERTYPES };

enum {NULL_HANDLER=0, TAPE_HANDLER, DISK_HANDLER, NET_HANDLER };

/* Task command list ... */

#define NULL_COMMAND            0
#define GO_COMMAND              1
#define STOP_COMMAND            2
#define SETPOS_COMMAND          3
#define NEWRUNFILE_COMMAND      4
#define REVOKERUNFILE_COMMAND   5
#define NEWDEVICE_COMMAND       6
#define SETPASS_COMMAND         6
#define RESTART_COMMAND         7
#define FLUSH_COMMAND           8
#define EXIT_COMMAND            9
#define LOAD_COMMAND           10
#define HANDLER_STATISTICS_COMMAND 11
#define PAUSE_COMMAND          12
#define RESUME_COMMAND         13
#define N_COMMANDS             14

#define USER_COMMAND           N_COMMANDS
#define MAX_USER_COMMANDS      11

/* Output style choices ... ...*/


#define CHANGE_ON_FULL        0
#define CHANGE_ON_INPUT       4

#define NAME_FOLLOWS_RUNFILES 0
#define NAME_FOLLOWS_INPUT    1

#define NAME_UNINDEXED        0
#define NAME_INDEXED          2

/* Shared memory task control structure ... */
/* Consists of a reserved space for shared control variables
   followed by the data block allocations ... */

#define CNTL_SHARED_MEMORY_KEY    1023

/* Semaphore array offsets (Solaris limit is 25 per key) */

#define MTSORT_SEM       (sort_cntl->mtsort_syssemkey)
#define MTSORT_IOSEM       (sort_cntl->mtsort_iosemkey)

#define SERIALISE_EXECUTION        0
#define CLOSE_DOWN_TODO            1
#define SORTLOG_FREE               2
#define INPUT_COMMAND_IN_PROGRESS  3
#define EVENT_COMMAND_IN_PROGRESS  4
#define OUTPUT_COMMAND_IN_PROGRESS 5
#define INPUT_COMMAND_ENDED        6
#define EVENT_COMMAND_ENDED        7
#define OUTPUT_COMMAND_ENDED       8
#define SCHEDULER_COMMAND_SENT     9
#define SCHEDULER_COMMAND_ENDED    10
#define SCHEDULER_RESPONSE_SENT    11
#define SCHEDULER_RESPONSE_ENDED   12

#define DISK_UPDATING_TODO         15
#define DISK_UPDATING_FREEBLOCKS   16
#define DISK_UPDATING_EXITED       17

#define MTSORT_SYS_SEM_ARRAY_LENGTH    (DISK_UPDATING_EXITED +1)

#define INPUT_BUFFER_ACCESS_SEMAPHORE 0
#define INPUT_STREAM_RUNFILE       1
#define EVENT_STREAM_RUNFILE       3
#define OUTPUT_STREAM_RUNFILE      5

#define INPUT_BUFFERS_FREE         7
#define INPUT_BUFFERS_POSTED       8
#define EVENT_BUFFERS_FREE         9
#define EVENT_BUFFERS_POSTED       10
#define OUTPUT_BUFFERS_FREE1       11
#define OUTPUT_BUFFERS_FREE2       12
#define OUTPUT_BUFFERS_FREE3       13
#define OUTPUT_BUFFERS_FREE4       14
#define OUTPUT_BUFFERS_POSTED1     15
#define OUTPUT_BUFFERS_POSTED2     16
#define OUTPUT_BUFFERS_POSTED3     17
#define OUTPUT_BUFFERS_POSTED4     18

#define MTSORT_IO_SEM_ARRAY_LENGTH    (OUTPUT_BUFFERS_POSTED4 +1)


struct sort_cntl_struct {
  int    version;

  int    input_trace;
  int    event_trace;
  int    output_trace;

  int    buffers_start;       /* byte offset to 1st buffer */
  int    output_style;        /* Output file=input tape name, or not ... */
  int    scheduler_pid;
  int    sleep_time;          /* msecs (used by net handler,...)  */

  int    input_static_handler_loaded;
  int    event_static_handler_loaded;
  int    output_static_handler_loaded;
  int    input_handler_loaded;      /* used during handler loading */
  int    event_handler_loaded;      /* used during handler loading */
  int    event_buffer_sync;           

  int    output_handler_loaded[MAX_OUTPUT_STREAMS];      /* used during handler loading */
  
  int    input_maxstreams;
  int    input_streams_inuse;

  int    event_maxstreams;
  int    event_streams_inuse;

  int    output_maxstreams;
  int    output_streams_inuse;
  int    output_blocklength;
  int    mtsort_syssemkey;

  int    mtsort_iosemkey;
  int    comms_port;
  int    control_using_memory;
  int    localdata_using_memory;

  long long    output_maxfileoffset;
  int    output_maxfile;
  int    event_pause;              /* 0= run normally; 1=pause before each event */

  /* Input handler data grouped together ... */

  struct input_stream_struct {
    int enabled;
    int stream;
    int stream_state;
    int handler;
    int command_action;
    int first_block;
    int last_block;
    pthread_t in_tid;
    pthread_t tr_tid;
    void   *current_buffer;
    void   *current_translator_buffer;
    struct datarec data_record;  /* Current data file info (256 bytes each) */
    int    nth_record;           /* Save of data record number */
    char   device[32];
    long long  blocks_processed;
  } input[MAX_INPUT_STREAMS];

  int    input_buffer_offset[MAX_INPUT_BUFFERS];

  int    input_data_length[MAX_INPUT_BUFFERS+1];
  int    input_data_offset[MAX_INPUT_BUFFERS+1];

  int    event_buffer_offset[MAX_INPUT_BUFFERS+1];
  int    event_data_length[MAX_INPUT_BUFFERS+1];

  /* Output handler data grouped together ... */

  struct output_stream_struct {
    int enabled;
    int stream;
    int stream_state;
    int handler;
    int command_action;
    int first_block;
    int last_block;
    pthread_t tid;
    char current_volume[12];
    char current_file[200];
    void *current_buffer;
    struct datarec data_record;  /* Current data file info (256 bytes each) */
    int  nth_record;           /* Save of data record number */
    char device[32];
    long long  blocks_processed;

    int output_buffer_offset[OUTPUT_BUFFERS_PER_STREAM+1];
    int output_data_length[OUTPUT_BUFFERS_PER_STREAM+1];
  } output[MAX_OUTPUT_STREAMS];

  /* Event handler pointers ... */

  struct event_stream_struct {
    int enabled;
    int stream;
    int stream_state;
    int command_action;
    int first_block;
    int last_block;
    pthread_t tid;
    char current_volume[12];
    char current_file[200];
    int current_buffer;
    int *eventstart;
    int *eventptr;
    int *eventend;
  } event[MAX_EVENT_STREAMS];

  int localdata_shmid;

  /* Data handler buffer transfer semaphores ... */

  void   *input_buffer_ready[MAX_INPUT_BUFFERS+1];
  int    input_buffer_ready_start,input_buffer_ready_end;
  void   *input_buffer_free[MAX_INPUT_BUFFERS+1];
  int    input_buffer_free_start,input_buffer_free_end;

  void   *event_buffer_ready[MAX_INPUT_BUFFERS+1];
  int    event_buffer_ready_start,event_buffer_ready_end;
  void   *event_buffer_free[MAX_INPUT_BUFFERS+1];
  int    event_buffer_free_start,event_buffer_free_end;

  void   *output_buffer_ready[MAX_OUTPUT_STREAMS][OUTPUT_BUFFERS_PER_STREAM+1];
  int    output_buffer_ready_start[MAX_OUTPUT_STREAMS];
  int    output_buffer_ready_end[MAX_OUTPUT_STREAMS];
  void   *output_buffer_free[MAX_OUTPUT_STREAMS][OUTPUT_BUFFERS_PER_STREAM+1];
  int    output_buffer_free_start[MAX_OUTPUT_STREAMS];
  int    output_buffer_free_end[MAX_OUTPUT_STREAMS];

  int    output_buffer_offset[MAX_OUTPUT_BUFFERS+1];
  int    output_data_length[MAX_OUTPUT_STREAMS][OUTPUT_BUFFERS_PER_STREAM+1];

  /* Commands from handlers to scheduler ... */

  char   scheduler_command_message[256];
  int    scheduler_command_source;

  /* Responses from scheduler to external command handler ... */

  char   scheduler_response_message[256];

  /* Commands from scheduler to specific handlers (using SIGPOLL) ... */

  int    input_command;
  int    input_command_stream;
  int    input_command_data;
  char   input_command_message[256];

  int    event_command;
  int    event_command_stream;
  int    event_command_data;
  char   event_command_message[256];

  int    output_command;
  int    output_command_stream;
  int    output_command_data;
  char   output_command_message[256];

  /* Variable part of Handlers ... */

  char input_loadname[128];
  char output_loadname[MAX_OUTPUT_STREAMS][128];
  char event_loadname[128];

  struct statistics_record event_statistics;

  char divider[128];
  char basedir[128];
  char sortdir[128];
  char specdir[128];
  char sortlog_file[128];

  /* Autogain runfile struct copy ... */

  struct headerrec1 agptr;
  int autogain_badpeak;

};

#define CNTL_SHARED_MEMORY_RESERVED_SIZE  (((sizeof(struct sort_cntl_struct) + 2047) >> 11) << 11)

#define CNTL_SHARED_MEMORY_SIZE  (CNTL_SHARED_MEMORY_RESERVED_SIZE +  \
		       (MAX_INPUT_BUFFERS*INPUT_BUFFER_ALLOCATED_SIZE)+ \
		       (MAX_EVENT_BUFFERS*EVENT_BUFFER_ALLOCATED_SIZE)+ \
		       (MAX_OUTPUT_BUFFERS*OUTPUT_BUFFER_ALLOCATED_SIZE))




/* Mapped file definitions ... */

typedef struct PATH_DESCR {
  char *path;
  int fd;
  char *mapping;   /* mapping address */
} PDESC;

/* Eurogam data block header format ... */

struct block_header {
  char name[8];
  int blockno;
  short EGdatastreamno;
  short streamno;
  int dummy;
  int nbytes;
};

/* New Tape Serverdata block header format ... */

typedef struct ts_block_header {
  char name[8];
  int blockno;
  short EGdatastreamno;
  short streamno;
  short TSEndian;
  short DataEndian;
  int nbytes;
} TSHeader;


/* Following are variables definitions, included only once ... */

#ifndef INCSORTCNTL

#ifdef _MTMAIN
volatile struct sort_cntl_struct *sort_cntl;
int sort_cntl_shmid;
void *localdata;

void *input_buffer_address[MAX_INPUT_BUFFERS];
void *event_buffer_address[MAX_EVENT_BUFFERS];
void *output_buffer_address[MAX_OUTPUT_BUFFERS];

int sortlog_OK;

char errmessage[128];

char basedir[128];

extern key_t key_template;

#else

extern volatile struct sort_cntl_struct *sort_cntl;
extern int sort_cntl_shmid;
extern void *localdata;

extern void *input_buffer_address[];
extern void *event_buffer_address[];
extern void *output_buffer_address[];

extern int sortlog_OK;

extern char basedir[];

#endif

#ifndef _SORTINPUT

#ifdef INPUT_HANDLER
extern unsigned short *inbuf, *inend, *inptr;
extern unsigned short *tranbuf, *tranend, *tranptr;
extern int input_record_needed[];
extern int input_eof_reached,input_eof_code;
extern int input_blocks_read;
extern int input_blocks_translated;
extern int goodblocks,badblocks;
extern int input_badevents_translated,input_goodevents_translated;
extern int badevents,goodevents;
extern int cz_compressed,gz_compressed;
#endif

#endif

/* Sort message definitions ...*/

static char messagebuf[128];

#define LOG_MESSAGE     0x00
#define LOG_ERROR       0x01
#define LOG_NOTIME      0x80

#define INCSORTCNTL
#endif



